@@ -128,7 +128,7 @@ int QgsLayoutItem::type() const
 
 QgsLayoutItem::Flags QgsLayoutItem::itemFlags() const
 {
-  return nullptr;
+  return QgsLayoutItem::Flags();
 }
 
 void QgsLayoutItem::setId( const QString &id )
@@ -342,13 +342,12 @@ void QgsLayoutItem::paint( QPainter *painter, const QStyleOptionGraphicsItem *it
     {
       // can reuse last cached image
       QgsRenderContext context = QgsLayoutUtils::createRenderContextForLayout( mLayout, painter, destinationDpi );
-      painter->save();
+      QgsScopedQPainterState painterState( painter );
       preparePainter( painter );
       double cacheScale = destinationDpi / mItemCacheDpi;
       painter->scale( cacheScale / context.scaleFactor(), cacheScale / context.scaleFactor() );
       painter->drawImage( boundingRect().x() * context.scaleFactor() / cacheScale,
                           boundingRect().y() * context.scaleFactor() / cacheScale, mItemCachedImage );
-      painter->restore();
       return;
     }
     else
@@ -379,12 +378,11 @@ void QgsLayoutItem::paint( QPainter *painter, const QStyleOptionGraphicsItem *it
 
       QgsImageOperation::multiplyOpacity( image, mEvaluatedOpacity );
 
-      painter->save();
+      QgsScopedQPainterState painterState( painter );
       // scale painter from mm to dots
       painter->scale( 1.0 / context.scaleFactor(), 1.0 / context.scaleFactor() );
       painter->drawImage( boundingRect().x() * context.scaleFactor(),
                           boundingRect().y() * context.scaleFactor(), image );
-      painter->restore();
 
       if ( previewRender )
       {
@@ -396,7 +394,7 @@ void QgsLayoutItem::paint( QPainter *painter, const QStyleOptionGraphicsItem *it
   else
   {
     // no caching or flattening
-    painter->save();
+    QgsScopedQPainterState painterState( painter );
     preparePainter( painter );
     QgsRenderContext context = QgsLayoutUtils::createRenderContextForLayout( mLayout, painter, destinationDpi );
     context.setExpressionContext( createExpressionContext() );
@@ -410,8 +408,6 @@ void QgsLayoutItem::paint( QPainter *painter, const QStyleOptionGraphicsItem *it
 
     painter->scale( context.scaleFactor(), context.scaleFactor() );
     drawFrame( context );
-
-    painter->restore();
   }
 }
 
@@ -580,8 +576,8 @@ void QgsLayoutItem::setScenePos( const QPointF destinationPos )
   //since setPos does not account for item rotation, use difference between
   //current scenePos (which DOES account for rotation) and destination pos
   //to calculate how much the item needs to move
-  if ( parentItem() )
-    setPos( pos() + ( destinationPos - scenePos() ) + parentItem()->scenePos() );
+  if ( auto *lParentItem = parentItem() )
+    setPos( pos() + ( destinationPos - scenePos() ) + lParentItem->scenePos() );
   else
     setPos( pos() + ( destinationPos - scenePos() ) );
 }
@@ -593,6 +589,9 @@ bool QgsLayoutItem::shouldBlockUndoCommands() const
 
 bool QgsLayoutItem::shouldDrawItem() const
 {
+  if ( mLayout && QgsLayoutUtils::itemIsAClippingSource( this ) )
+    return false;
+
   if ( !mLayout || mLayout->renderContext().isPreviewRender() )
   {
     //preview mode so OK to draw item
@@ -691,7 +690,7 @@ bool QgsLayoutItem::writeXml( QDomElement &parentElement, QDomDocument &doc, con
 
 bool QgsLayoutItem::readXml( const QDomElement &element, const QDomDocument &doc, const QgsReadWriteContext &context )
 {
-  if ( element.nodeName() != QStringLiteral( "LayoutItem" ) )
+  if ( element.nodeName() != QLatin1String( "LayoutItem" ) )
   {
     return false;
   }
@@ -1169,6 +1168,11 @@ bool QgsLayoutItem::accept( QgsStyleEntityVisitorInterface *visitor ) const
   return true;
 }
 
+QgsGeometry QgsLayoutItem::clipPath() const
+{
+  return QgsGeometry();
+}
+
 void QgsLayoutItem::refresh()
 {
   QgsLayoutObject::refresh();
@@ -1199,12 +1203,18 @@ void QgsLayoutItem::drawDebugRect( QPainter *painter )
     return;
   }
 
-  painter->save();
+  QgsScopedQPainterState painterState( painter );
   painter->setRenderHint( QPainter::Antialiasing, false );
   painter->setPen( Qt::NoPen );
   painter->setBrush( QColor( 100, 255, 100, 200 ) );
   painter->drawRect( rect() );
-  painter->restore();
+}
+
+QPainterPath QgsLayoutItem::framePath() const
+{
+  QPainterPath path;
+  path.addRect( QRectF( 0, 0, rect().width(), rect().height() ) );
+  return path;
 }
 
 void QgsLayoutItem::drawFrame( QgsRenderContext &context )
@@ -1213,11 +1223,14 @@ void QgsLayoutItem::drawFrame( QgsRenderContext &context )
     return;
 
   QPainter *p = context.painter();
-  p->save();
+
+  QgsScopedQPainterState painterState( p );
+
   p->setPen( pen() );
   p->setBrush( Qt::NoBrush );
-  p->drawRect( QRectF( 0, 0, rect().width(), rect().height() ) );
-  p->restore();
+  context.setPainterFlagsUsingContext( p );
+
+  p->drawPath( framePath() );
 }
 
 void QgsLayoutItem::drawBackground( QgsRenderContext &context )
@@ -1225,12 +1238,14 @@ void QgsLayoutItem::drawBackground( QgsRenderContext &context )
   if ( !mBackground || !context.painter() )
     return;
 
+  QgsScopedQPainterState painterState( context.painter() );
+
   QPainter *p = context.painter();
-  p->save();
   p->setBrush( brush() );
   p->setPen( Qt::NoPen );
-  p->drawRect( QRectF( 0, 0, rect().width(), rect().height() ) );
-  p->restore();
+  context.setPainterFlagsUsingContext( p );
+
+  p->drawPath( framePath() );
 }
 
 void QgsLayoutItem::setFixedSize( const QgsLayoutSize &size )
@@ -1333,6 +1348,10 @@ void QgsLayoutItem::preparePainter( QPainter *painter )
   }
 
   painter->setRenderHint( QPainter::Antialiasing, shouldDrawAntialiased() );
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 13, 0)
+  painter->setRenderHint( QPainter::LosslessImageRendering, mLayout && mLayout->renderContext().testFlag( QgsLayoutRenderContext::FlagLosslessImageRendering ) );
+#endif
 }
 
 bool QgsLayoutItem::shouldDrawAntialiased() const
