@@ -33,6 +33,7 @@
 #include "qgsmeshlayer.h"
 #include "qgsreferencedgeometry.h"
 #include "qgsrasterfilewriter.h"
+#include "qgsvectortilelayer.h"
 
 QList<QgsRasterLayer *> QgsProcessingUtils::compatibleRasterLayers( QgsProject *project, bool sort )
 {
@@ -133,6 +134,25 @@ QList<QgsMapLayer *> QgsProcessingUtils::compatibleLayers( QgsProject *project,
   return layers;
 }
 
+QString QgsProcessingUtils::encodeProviderKeyAndUri( const QString &providerKey, const QString &uri )
+{
+  return QStringLiteral( "%1://%2" ).arg( providerKey, uri );
+}
+
+bool QgsProcessingUtils::decodeProviderKeyAndUri( const QString &string, QString &providerKey, QString &uri )
+{
+  QRegularExpression re( QStringLiteral( "^(\\w+?):\\/\\/(.+)$" ) );
+  const QRegularExpressionMatch match = re.match( string );
+  if ( !match.hasMatch() )
+    return false;
+
+  providerKey = match.captured( 1 );
+  uri = match.captured( 2 );
+
+  // double check that provider is valid
+  return QgsProviderRegistry::instance()->providerMetadata( providerKey );
+}
+
 QgsMapLayer *QgsProcessingUtils::mapLayerFromStore( const QString &string, QgsMapLayerStore *store, QgsProcessingUtils::LayerHint typeHint )
 {
   if ( !store || string.isEmpty() )
@@ -152,6 +172,10 @@ QgsMapLayer *QgsProcessingUtils::mapLayerFromStore( const QString &string, QgsMa
         return true;
       case QgsMapLayerType::MeshLayer:
         return !canUseLayer( qobject_cast< QgsMeshLayer * >( layer ) );
+      case QgsMapLayerType::VectorTileLayer:
+        return !canUseLayer( qobject_cast< QgsVectorTileLayer * >( layer ) );
+      case QgsMapLayerType::AnnotationLayer:
+        return true;
     }
     return true;
   } ), layers.end() );
@@ -195,19 +219,33 @@ QgsMapLayer *QgsProcessingUtils::mapLayerFromStore( const QString &string, QgsMa
 
 QgsMapLayer *QgsProcessingUtils::loadMapLayerFromString( const QString &string, const QgsCoordinateTransformContext &transformContext, LayerHint typeHint )
 {
-  QStringList components = string.split( '|' );
-  if ( components.isEmpty() )
-    return nullptr;
+  QString provider;
+  QString uri;
+  const bool useProvider = decodeProviderKeyAndUri( string, provider, uri );
+  if ( !useProvider )
+    uri = string;
 
-  QFileInfo fi;
-  if ( QFileInfo::exists( string ) )
-    fi = QFileInfo( string );
-  else if ( QFileInfo::exists( components.at( 0 ) ) )
-    fi = QFileInfo( components.at( 0 ) );
+  QString name;
+  // for disk based sources, we use the filename to determine a layer name
+  if ( !useProvider || ( provider == QLatin1String( "ogr" ) || provider == QLatin1String( "gdal" ) || provider == QLatin1String( "mdal" ) ) )
+  {
+    QStringList components = uri.split( '|' );
+    if ( components.isEmpty() )
+      return nullptr;
+
+    QFileInfo fi;
+    if ( QFileInfo::exists( uri ) )
+      fi = QFileInfo( uri );
+    else if ( QFileInfo::exists( components.at( 0 ) ) )
+      fi = QFileInfo( components.at( 0 ) );
+    else
+      return nullptr;
+    name = fi.baseName();
+  }
   else
-    return nullptr;
-
-  QString name = fi.baseName();
+  {
+    name = QgsDataSourceUri( uri ).table();
+  }
 
   // brute force attempt to load a matching layer
   if ( typeHint == LayerHint::UnknownType || typeHint == LayerHint::Vector )
@@ -215,7 +253,17 @@ QgsMapLayer *QgsProcessingUtils::loadMapLayerFromString( const QString &string,
     QgsVectorLayer::LayerOptions options { transformContext };
     options.loadDefaultStyle = false;
     options.skipCrsValidation = true;
-    std::unique_ptr< QgsVectorLayer > layer = qgis::make_unique<QgsVectorLayer>( string, name, QStringLiteral( "ogr" ), options );
+
+    std::unique_ptr< QgsVectorLayer > layer;
+    if ( useProvider )
+    {
+      layer = qgis::make_unique<QgsVectorLayer>( uri, name, provider, options );
+    }
+    else
+    {
+      // fallback to ogr
+      layer = qgis::make_unique<QgsVectorLayer>( uri, name, QStringLiteral( "ogr" ), options );
+    }
     if ( layer->isValid() )
     {
       return layer.release();
@@ -226,7 +274,18 @@ QgsMapLayer *QgsProcessingUtils::loadMapLayerFromString( const QString &string,
     QgsRasterLayer::LayerOptions rasterOptions;
     rasterOptions.loadDefaultStyle = false;
     rasterOptions.skipCrsValidation = true;
-    std::unique_ptr< QgsRasterLayer > rasterLayer( new QgsRasterLayer( string, name, QStringLiteral( "gdal" ), rasterOptions ) );
+
+    std::unique_ptr< QgsRasterLayer > rasterLayer;
+    if ( useProvider )
+    {
+      rasterLayer = qgis::make_unique< QgsRasterLayer >( uri, name, provider, rasterOptions );
+    }
+    else
+    {
+      // fallback to gdal
+      rasterLayer = qgis::make_unique< QgsRasterLayer >( uri, name, QStringLiteral( "gdal" ), rasterOptions );
+    }
+
     if ( rasterLayer->isValid() )
     {
       return rasterLayer.release();
@@ -236,7 +295,16 @@ QgsMapLayer *QgsProcessingUtils::loadMapLayerFromString( const QString &string,
   {
     QgsMeshLayer::LayerOptions meshOptions;
     meshOptions.skipCrsValidation = true;
-    std::unique_ptr< QgsMeshLayer > meshLayer( new QgsMeshLayer( string, name, QStringLiteral( "mdal" ), meshOptions ) );
+
+    std::unique_ptr< QgsMeshLayer > meshLayer;
+    if ( useProvider )
+    {
+      meshLayer = qgis::make_unique< QgsMeshLayer >( uri, name, provider, meshOptions );
+    }
+    else
+    {
+      meshLayer = qgis::make_unique< QgsMeshLayer >( uri, name, QStringLiteral( "mdal" ), meshOptions );
+    }
     if ( meshLayer->isValid() )
     {
       return meshLayer.release();
@@ -252,9 +320,9 @@ QgsMapLayer *QgsProcessingUtils::mapLayerFromString( const QString &string, QgsP
 
   // prefer project layers
   QgsMapLayer *layer = nullptr;
-  if ( context.project() )
+  if ( auto *lProject = context.project() )
   {
-    QgsMapLayer *layer = mapLayerFromStore( string, context.project()->layerStore(), typeHint );
+    QgsMapLayer *layer = mapLayerFromStore( string, lProject->layerStore(), typeHint );
     if ( layer )
       return layer;
   }
@@ -282,12 +350,18 @@ QgsProcessingFeatureSource *QgsProcessingUtils::variantToSource( const QVariant
 {
   QVariant val = value;
   bool selectedFeaturesOnly = false;
+  long long featureLimit = -1;
+  bool overrideGeometryCheck = false;
+  QgsFeatureRequest::InvalidGeometryCheck geometryCheck = QgsFeatureRequest::GeometryAbortOnInvalid;
   if ( val.canConvert<QgsProcessingFeatureSourceDefinition>() )
   {
     // input is a QgsProcessingFeatureSourceDefinition - get extra properties from it
     QgsProcessingFeatureSourceDefinition fromVar = qvariant_cast<QgsProcessingFeatureSourceDefinition>( val );
     selectedFeaturesOnly = fromVar.selectedFeaturesOnly;
+    featureLimit = fromVar.featureLimit;
     val = fromVar.source;
+    overrideGeometryCheck = fromVar.flags & QgsProcessingFeatureSourceDefinition::Flag::FlagOverrideDefaultGeometryCheck;
+    geometryCheck = fromVar.geometryCheck;
   }
   else if ( val.canConvert<QgsProcessingOutputLayerDefinition>() )
   {
@@ -298,7 +372,10 @@ QgsProcessingFeatureSource *QgsProcessingUtils::variantToSource( const QVariant
 
   if ( QgsVectorLayer *layer = qobject_cast< QgsVectorLayer * >( qvariant_cast<QObject *>( val ) ) )
   {
-    return new QgsProcessingFeatureSource( layer, context );
+    std::unique_ptr< QgsProcessingFeatureSource> source = qgis::make_unique< QgsProcessingFeatureSource >( layer, context, false, featureLimit );
+    if ( overrideGeometryCheck )
+      source->setInvalidGeometryCheck( geometryCheck );
+    return source.release();
   }
 
   QString layerRef;
@@ -311,7 +388,10 @@ QgsProcessingFeatureSource *QgsProcessingUtils::variantToSource( const QVariant
     // fall back to default
     if ( QgsVectorLayer *layer = qobject_cast< QgsVectorLayer * >( qvariant_cast<QObject *>( fallbackValue ) ) )
     {
-      return new QgsProcessingFeatureSource( layer, context );
+      std::unique_ptr< QgsProcessingFeatureSource> source = qgis::make_unique< QgsProcessingFeatureSource >( layer, context, false, featureLimit );
+      if ( overrideGeometryCheck )
+        source->setInvalidGeometryCheck( geometryCheck );
+      return source.release();
     }
 
     layerRef = fallbackValue.toString();
@@ -328,14 +408,80 @@ QgsProcessingFeatureSource *QgsProcessingUtils::variantToSource( const QVariant
   if ( !vl )
     return nullptr;
 
+  std::unique_ptr< QgsProcessingFeatureSource> source;
   if ( selectedFeaturesOnly )
   {
-    return new QgsProcessingFeatureSource( new QgsVectorLayerSelectedFeatureSource( vl ), context, true );
+    source = qgis::make_unique< QgsProcessingFeatureSource>( new QgsVectorLayerSelectedFeatureSource( vl ), context, true, featureLimit );
   }
   else
   {
-    return new QgsProcessingFeatureSource( vl, context );
+    source = qgis::make_unique< QgsProcessingFeatureSource >( vl, context, false, featureLimit );
   }
+
+  if ( overrideGeometryCheck )
+    source->setInvalidGeometryCheck( geometryCheck );
+  return source.release();
+}
+
+QgsCoordinateReferenceSystem QgsProcessingUtils::variantToCrs( const QVariant &value, QgsProcessingContext &context, const QVariant &fallbackValue )
+{
+  QVariant val = value;
+
+  if ( val.canConvert<QgsCoordinateReferenceSystem>() )
+  {
+    // input is a QgsCoordinateReferenceSystem - done!
+    return val.value< QgsCoordinateReferenceSystem >();
+  }
+  else if ( val.canConvert<QgsProcessingFeatureSourceDefinition>() )
+  {
+    // input is a QgsProcessingFeatureSourceDefinition - get extra properties from it
+    QgsProcessingFeatureSourceDefinition fromVar = qvariant_cast<QgsProcessingFeatureSourceDefinition>( val );
+    val = fromVar.source;
+  }
+  else if ( val.canConvert<QgsProcessingOutputLayerDefinition>() )
+  {
+    // input is a QgsProcessingOutputLayerDefinition - get extra properties from it
+    QgsProcessingOutputLayerDefinition fromVar = qvariant_cast<QgsProcessingOutputLayerDefinition>( val );
+    val = fromVar.sink;
+  }
+
+  if ( val.canConvert<QgsProperty>() && val.value< QgsProperty >().propertyType() == QgsProperty::StaticProperty )
+  {
+    val = val.value< QgsProperty >().staticValue();
+  }
+
+  // maybe a map layer
+  if ( QgsMapLayer *layer = qobject_cast< QgsMapLayer * >( qvariant_cast<QObject *>( val ) ) )
+    return layer->crs();
+
+  if ( val.canConvert<QgsProperty>() )
+    val = val.value< QgsProperty >().valueAsString( context.expressionContext(), fallbackValue.toString() );
+
+  if ( !val.isValid() )
+  {
+    // fall back to default
+    val = fallbackValue;
+  }
+
+  QString crsText = val.toString();
+  if ( crsText.isEmpty() )
+    crsText = fallbackValue.toString();
+
+  if ( crsText.isEmpty() )
+    return QgsCoordinateReferenceSystem();
+
+  // maybe special string
+  if ( context.project() && crsText.compare( QLatin1String( "ProjectCrs" ), Qt::CaseInsensitive ) == 0 )
+    return context.project()->crs();
+
+  // maybe a map layer reference
+  if ( QgsMapLayer *layer = QgsProcessingUtils::mapLayerFromString( crsText, context ) )
+    return layer->crs();
+
+  // else CRS from string
+  QgsCoordinateReferenceSystem crs;
+  crs.createFromString( crsText );
+  return crs;
 }
 
 bool QgsProcessingUtils::canUseLayer( const QgsMeshLayer *layer )
@@ -343,6 +489,11 @@ bool QgsProcessingUtils::canUseLayer( const QgsMeshLayer *layer )
   return layer && layer->dataProvider();
 }
 
+bool QgsProcessingUtils::canUseLayer( const QgsVectorTileLayer *layer )
+{
+  return layer && layer->isValid();
+}
+
 bool QgsProcessingUtils::canUseLayer( const QgsRasterLayer *layer )
 {
   return layer && layer->isValid();
@@ -438,6 +589,18 @@ QString QgsProcessingUtils::variantToPythonLiteral( const QVariant &value )
       return parts.join( ',' ).prepend( '[' ).append( ']' );
     }
 
+    case QVariant::Map:
+    {
+      const QVariantMap map = value.toMap();
+      QStringList parts;
+      parts.reserve( map.size() );
+      for ( auto it = map.constBegin(); it != map.constEnd(); ++it )
+      {
+        parts << QStringLiteral( "%1: %2" ).arg( stringToPythonLiteral( it.key() ), variantToPythonLiteral( it.value() ) );
+      }
+      return parts.join( ',' ).prepend( '{' ).append( '}' );
+    }
+
     default:
       break;
   }
@@ -448,12 +611,12 @@ QString QgsProcessingUtils::variantToPythonLiteral( const QVariant &value )
 QString QgsProcessingUtils::stringToPythonLiteral( const QString &string )
 {
   QString s = string;
-  s.replace( '\\', QStringLiteral( "\\\\" ) );
-  s.replace( '\n', QStringLiteral( "\\n" ) );
-  s.replace( '\r', QStringLiteral( "\\r" ) );
-  s.replace( '\t', QStringLiteral( "\\t" ) );
-  s.replace( '"', QStringLiteral( "\\\"" ) );
-  s.replace( '\'', QStringLiteral( "\\\'" ) );
+  s.replace( '\\', QLatin1String( "\\\\" ) );
+  s.replace( '\n', QLatin1String( "\\n" ) );
+  s.replace( '\r', QLatin1String( "\\r" ) );
+  s.replace( '\t', QLatin1String( "\\t" ) );
+  s.replace( '"', QLatin1String( "\\\"" ) );
+  s.replace( '\'', QLatin1String( "\\\'" ) );
   s = s.prepend( '\'' ).append( '\'' );
   return s;
 }
@@ -461,16 +624,26 @@ QString QgsProcessingUtils::stringToPythonLiteral( const QString &string )
 void QgsProcessingUtils::parseDestinationString( QString &destination, QString &providerKey, QString &uri, QString &layerName, QString &format, QMap<QString, QVariant> &options, bool &useWriter, QString &extension )
 {
   extension.clear();
-  QRegularExpression splitRx( QStringLiteral( "^(.{3,}?):(.*)$" ) );
-  QRegularExpressionMatch match = splitRx.match( destination );
-  if ( match.hasMatch() )
+  bool matched = decodeProviderKeyAndUri( destination, providerKey, uri );
+
+  if ( !matched )
   {
-    providerKey = match.captured( 1 );
-    if ( providerKey == QStringLiteral( "postgis" ) ) // older processing used "postgis" instead of "postgres"
+    QRegularExpression splitRx( QStringLiteral( "^(.{3,}?):(.*)$" ) );
+    QRegularExpressionMatch match = splitRx.match( destination );
+    if ( match.hasMatch() )
+    {
+      providerKey = match.captured( 1 );
+      uri = match.captured( 2 );
+      matched = true;
+    }
+  }
+
+  if ( matched )
+  {
+    if ( providerKey == QLatin1String( "postgis" ) ) // older processing used "postgis" instead of "postgres"
     {
       providerKey = QStringLiteral( "postgres" );
     }
-    uri = match.captured( 2 );
     if ( providerKey == QLatin1String( "ogr" ) )
     {
       QgsDataSourceUri dsUri( uri );
@@ -484,10 +657,12 @@ void QgsProcessingUtils::parseDestinationString( QString &destination, QString &
         uri = dsUri.database();
         extension = QFileInfo( uri ).completeSuffix();
         format = QgsVectorFileWriter::driverForExtension( extension );
+        options.insert( QStringLiteral( "driverName" ), format );
       }
       else
       {
         extension = QFileInfo( uri ).completeSuffix();
+        options.insert( QStringLiteral( "driverName" ), QgsVectorFileWriter::driverForExtension( extension ) );
       }
       options.insert( QStringLiteral( "update" ), true );
     }
@@ -497,6 +672,7 @@ void QgsProcessingUtils::parseDestinationString( QString &destination, QString &
   {
     useWriter = true;
     providerKey = QStringLiteral( "ogr" );
+
     QRegularExpression splitRx( QStringLiteral( "^(.*)\\.(.*?)$" ) );
     QRegularExpressionMatch match = splitRx.match( destination );
     if ( match.hasMatch() )
@@ -516,7 +692,7 @@ void QgsProcessingUtils::parseDestinationString( QString &destination, QString &
   }
 }
 
-QgsFeatureSink *QgsProcessingUtils::createFeatureSink( QString &destination, QgsProcessingContext &context, const QgsFields &fields, QgsWkbTypes::Type geometryType, const QgsCoordinateReferenceSystem &crs, const QVariantMap &createOptions, QgsFeatureSink::SinkFlags sinkFlags )
+QgsFeatureSink *QgsProcessingUtils::createFeatureSink( QString &destination, QgsProcessingContext &context, const QgsFields &fields, QgsWkbTypes::Type geometryType, const QgsCoordinateReferenceSystem &crs, const QVariantMap &createOptions, const QStringList &datasourceOptions, const QStringList &layerOptions, QgsFeatureSink::SinkFlags sinkFlags, QgsRemappingSinkDefinition *remappingDefinition )
 {
   QVariantMap options = createOptions;
   if ( !options.contains( QStringLiteral( "fileEncoding" ) ) )
@@ -541,6 +717,8 @@ QgsFeatureSink *QgsProcessingUtils::createFeatureSink( QString &destination, Qgs
       throw QgsProcessingException( QObject::tr( "Could not create memory layer" ) );
     }
 
+    layer->setCustomProperty( QStringLiteral( "OnConvertFormatRegeneratePrimaryKey" ), static_cast< bool >( sinkFlags & QgsFeatureSink::RegeneratePrimaryKey ) );
+
     // update destination to layer ID
     destination = layer->id();
 
@@ -566,40 +744,102 @@ QgsFeatureSink *QgsProcessingUtils::createFeatureSink( QString &destination, Qgs
       // use QgsVectorFileWriter for OGR destinations instead of QgsVectorLayerImport, as that allows
       // us to use any OGR format which supports feature addition
       QString finalFileName;
+      QString finalLayerName;
       QgsVectorFileWriter::SaveVectorOptions saveOptions;
       saveOptions.fileEncoding = options.value( QStringLiteral( "fileEncoding" ) ).toString();
+      saveOptions.layerName = !layerName.isEmpty() ? layerName : options.value( QStringLiteral( "layerName" ) ).toString();
       saveOptions.driverName = format;
-      saveOptions.datasourceOptions = QgsVectorFileWriter::defaultDatasetOptions( format );
-      saveOptions.layerOptions = QgsVectorFileWriter::defaultLayerOptions( format );
+      saveOptions.datasourceOptions = !datasourceOptions.isEmpty() ? datasourceOptions : QgsVectorFileWriter::defaultDatasetOptions( format );
+      saveOptions.layerOptions = !layerOptions.isEmpty() ? layerOptions : QgsVectorFileWriter::defaultLayerOptions( format );
       saveOptions.symbologyExport = QgsVectorFileWriter::NoSymbology;
-      saveOptions.actionOnExistingFile = QgsVectorFileWriter::CreateOrOverwriteFile;
-      std::unique_ptr< QgsVectorFileWriter > writer( QgsVectorFileWriter::create( destination, newFields, geometryType, crs, context.transformContext(), saveOptions, sinkFlags, &finalFileName ) );
+      if ( remappingDefinition )
+      {
+        saveOptions.actionOnExistingFile = QgsVectorFileWriter::AppendToLayerNoNewFields;
+        // sniff destination file to get correct wkb type and crs
+        std::unique_ptr< QgsVectorLayer > vl = qgis::make_unique< QgsVectorLayer >( destination );
+        if ( vl->isValid() )
+        {
+          remappingDefinition->setDestinationWkbType( vl->wkbType() );
+          remappingDefinition->setDestinationCrs( vl->crs() );
+          newFields = vl->fields();
+          remappingDefinition->setDestinationFields( newFields );
+        }
+        context.expressionContext().setFields( fields );
+      }
+      else
+      {
+        saveOptions.actionOnExistingFile = QgsVectorFileWriter::CreateOrOverwriteFile;
+      }
+      std::unique_ptr< QgsVectorFileWriter > writer( QgsVectorFileWriter::create( destination, newFields, geometryType, crs, context.transformContext(), saveOptions, sinkFlags, &finalFileName, &finalLayerName ) );
       if ( writer->hasError() )
       {
         throw QgsProcessingException( QObject::tr( "Could not create layer %1: %2" ).arg( destination, writer->errorMessage() ) );
       }
       destination = finalFileName;
-      return new QgsProcessingFeatureSink( writer.release(), destination, context, true );
+      if ( !saveOptions.layerName.isEmpty() && !finalLayerName.isEmpty() )
+        destination += QStringLiteral( "|layername=%1" ).arg( finalLayerName );
+
+      if ( remappingDefinition )
+      {
+        std::unique_ptr< QgsRemappingProxyFeatureSink > remapSink = qgis::make_unique< QgsRemappingProxyFeatureSink >( *remappingDefinition, writer.release(), true );
+        remapSink->setExpressionContext( context.expressionContext() );
+        remapSink->setTransformContext( context.transformContext() );
+        return new QgsProcessingFeatureSink( remapSink.release(), destination, context, true );
+      }
+      else
+        return new QgsProcessingFeatureSink( writer.release(), destination, context, true );
     }
     else
     {
-      //create empty layer
       const QgsVectorLayer::LayerOptions layerOptions { context.transformContext() };
-      std::unique_ptr< QgsVectorLayerExporter > exporter = qgis::make_unique<QgsVectorLayerExporter>( uri, providerKey, newFields, geometryType, crs, true, options, sinkFlags );
-      if ( exporter->errorCode() )
+      if ( remappingDefinition )
       {
-        throw QgsProcessingException( QObject::tr( "Could not create layer %1: %2" ).arg( destination, exporter->errorMessage() ) );
+        //write to existing layer
+
+        // use destination string as layer name (eg "postgis:..." )
+        if ( !layerName.isEmpty() )
+        {
+          QVariantMap parts = QgsProviderRegistry::instance()->decodeUri( providerKey, uri );
+          parts.insert( QStringLiteral( "layerName" ), layerName );
+          uri = QgsProviderRegistry::instance()->encodeUri( providerKey, parts );
+        }
+
+        std::unique_ptr< QgsVectorLayer > layer = qgis::make_unique<QgsVectorLayer>( uri, destination, providerKey, layerOptions );
+        // update destination to layer ID
+        destination = layer->id();
+        if ( layer->isValid() )
+        {
+          remappingDefinition->setDestinationWkbType( layer->wkbType() );
+          remappingDefinition->setDestinationCrs( layer->crs() );
+          remappingDefinition->setDestinationFields( layer->fields() );
+        }
+
+        std::unique_ptr< QgsRemappingProxyFeatureSink > remapSink = qgis::make_unique< QgsRemappingProxyFeatureSink >( *remappingDefinition, layer->dataProvider(), false );
+        context.temporaryLayerStore()->addMapLayer( layer.release() );
+        remapSink->setExpressionContext( context.expressionContext() );
+        remapSink->setTransformContext( context.transformContext() );
+        context.expressionContext().setFields( fields );
+        return new QgsProcessingFeatureSink( remapSink.release(), destination, context, true );
       }
+      else
+      {
+        //create empty layer
+        std::unique_ptr< QgsVectorLayerExporter > exporter = qgis::make_unique<QgsVectorLayerExporter>( uri, providerKey, newFields, geometryType, crs, true, options, sinkFlags );
+        if ( exporter->errorCode() )
+        {
+          throw QgsProcessingException( QObject::tr( "Could not create layer %1: %2" ).arg( destination, exporter->errorMessage() ) );
+        }
 
-      // use destination string as layer name (eg "postgis:..." )
-      if ( !layerName.isEmpty() )
-        uri += QStringLiteral( "|layername=%1" ).arg( layerName );
-      std::unique_ptr< QgsVectorLayer > layer = qgis::make_unique<QgsVectorLayer>( uri, destination, providerKey, layerOptions );
-      // update destination to layer ID
-      destination = layer->id();
+        // use destination string as layer name (eg "postgis:..." )
+        if ( !layerName.isEmpty() )
+        {
+          uri += QStringLiteral( "|layername=%1" ).arg( layerName );
+          // update destination to generated URI
+          destination = uri;
+        }
 
-      context.temporaryLayerStore()->addMapLayer( layer.release() );
-      return new QgsProcessingFeatureSink( exporter.release(), destination, context, true );
+        return new QgsProcessingFeatureSink( exporter.release(), destination, context, true );
+      }
     }
   }
 }
@@ -676,31 +916,50 @@ QVariant QgsProcessingUtils::generateIteratingDestination( const QVariant &input
     }
     else if ( res.startsWith( QLatin1String( "memory:" ) ) )
     {
-      return res + '_' + id.toString();
+      return QString( res + '_' + id.toString() );
     }
     else
     {
       // assume a filename type output for now
       // TODO - uris?
       int lastIndex = res.lastIndexOf( '.' );
-      return res.left( lastIndex ) + '_' + id.toString() + res.mid( lastIndex );
+      return QString( res.left( lastIndex ) + '_' + id.toString() + res.mid( lastIndex ) );
     }
   }
 }
 
 QString QgsProcessingUtils::tempFolder()
 {
+  // we maintain a list of temporary folders -- this allows us to append additional
+  // folders when a setting change causes the base temp folder to change, while deferring
+  // cleanup of ALL these temp folders until session end (we can't cleanup older folders immediately,
+  // because we don't know whether they have data in them which is still wanted)
+  static std::vector< std::unique_ptr< QTemporaryDir > > sTempFolders;
   static QString sFolder;
   static QMutex sMutex;
-  sMutex.lock();
-  if ( sFolder.isEmpty() || !sFolder.startsWith( QgsSettings().value( QStringLiteral( "Processing/Configuration/TEMP_PATH" ), QDir::tempPath() ).toString() ) )
+  QMutexLocker locker( &sMutex );
+  const QString basePath = QgsSettings().value( QStringLiteral( "Processing/Configuration/TEMP_PATH2" ) ).toString();
+  if ( basePath.isEmpty() )
   {
-    QString subPath = QUuid::createUuid().toString().remove( '-' ).remove( '{' ).remove( '}' );
-    sFolder = QgsSettings().value( QStringLiteral( "Processing/Configuration/TEMP_PATH" ), QDir::tempPath() ).toString() + QStringLiteral( "/processing_" ) + subPath;
-    if ( !QDir( sFolder ).exists() )
-      QDir().mkpath( sFolder );
+    // default setting -- automatically create a temp folder
+    if ( sTempFolders.empty() )
+    {
+      const QString templatePath = QStringLiteral( "%1/processing_XXXXXX" ).arg( QDir::tempPath() );
+      std::unique_ptr< QTemporaryDir > tempFolder = qgis::make_unique< QTemporaryDir >( templatePath );
+      sFolder = tempFolder->path();
+      sTempFolders.emplace_back( std::move( tempFolder ) );
+    }
+  }
+  else if ( sFolder.isEmpty() || !sFolder.startsWith( basePath ) || sTempFolders.empty() )
+  {
+    if ( !QDir().exists( basePath ) )
+      QDir().mkpath( basePath );
+
+    const QString templatePath = QStringLiteral( "%1/processing_XXXXXX" ).arg( basePath );
+    std::unique_ptr< QTemporaryDir > tempFolder = qgis::make_unique< QTemporaryDir >( templatePath );
+    sFolder = tempFolder->path();
+    sTempFolders.emplace_back( std::move( tempFolder ) );
   }
-  sMutex.unlock();
   return sFolder;
 }
 
@@ -757,11 +1016,12 @@ QString QgsProcessingUtils::formatHelpMapAsHtml( const QVariantMap &map, const Q
   if ( !map.value( QStringLiteral( "ALG_VERSION" ) ).toString().isEmpty() )
     s += QObject::tr( "<p align=\"right\">Algorithm version: %1</p>" ).arg( getText( QStringLiteral( "ALG_VERSION" ) ) );
 
-  s += QStringLiteral( "</body></html>" );
+  s += QLatin1String( "</body></html>" );
   return s;
 }
 
-QString convertToCompatibleFormatInternal( const QgsVectorLayer *vl, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback, QString *layerName )
+QString convertToCompatibleFormatInternal( const QgsVectorLayer *vl, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback, QString *layerName,
+    long long featureLimit )
 {
   bool requiresTranslation = false;
 
@@ -769,6 +1029,9 @@ QString convertToCompatibleFormatInternal( const QgsVectorLayer *vl, bool select
   // as we need to subset only selected features, a concept which doesn't exist outside QGIS!
   requiresTranslation = requiresTranslation || selectedFeaturesOnly;
 
+  // if we are limiting the feature count, we better export
+  requiresTranslation = requiresTranslation || featureLimit != -1;
+
   // if the data provider is NOT ogr, then we HAVE to convert. Otherwise we run into
   // issues with data providers like spatialite, delimited text where the format can be
   // opened outside of QGIS, but with potentially very different behavior!
@@ -823,10 +1086,20 @@ QString convertToCompatibleFormatInternal( const QgsVectorLayer *vl, bool select
     std::unique_ptr< QgsVectorFileWriter > writer( QgsVectorFileWriter::create( temp, vl->fields(), vl->wkbType(), vl->crs(), context.transformContext(), saveOptions ) );
     QgsFeature f;
     QgsFeatureIterator it;
-    if ( selectedFeaturesOnly )
-      it = vl->getSelectedFeatures();
+    if ( featureLimit != -1 )
+    {
+      if ( selectedFeaturesOnly )
+        it = vl->getSelectedFeatures( QgsFeatureRequest().setLimit( featureLimit ) );
+      else
+        it = vl->getFeatures( QgsFeatureRequest().setLimit( featureLimit ) );
+    }
     else
-      it = vl->getFeatures();
+    {
+      if ( selectedFeaturesOnly )
+        it = vl->getSelectedFeatures( QgsFeatureRequest().setLimit( featureLimit ) );
+      else
+        it = vl->getFeatures();
+    }
 
     while ( it.nextFeature( f ) )
     {
@@ -842,15 +1115,15 @@ QString convertToCompatibleFormatInternal( const QgsVectorLayer *vl, bool select
   }
 }
 
-QString QgsProcessingUtils::convertToCompatibleFormat( const QgsVectorLayer *vl, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback )
+QString QgsProcessingUtils::convertToCompatibleFormat( const QgsVectorLayer *vl, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback, long long featureLimit )
 {
-  return convertToCompatibleFormatInternal( vl, selectedFeaturesOnly, baseName, compatibleFormats, preferredFormat, context, feedback, nullptr );
+  return convertToCompatibleFormatInternal( vl, selectedFeaturesOnly, baseName, compatibleFormats, preferredFormat, context, feedback, nullptr, featureLimit );
 }
 
-QString QgsProcessingUtils::convertToCompatibleFormatAndLayerName( const QgsVectorLayer *layer, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback, QString &layerName )
+QString QgsProcessingUtils::convertToCompatibleFormatAndLayerName( const QgsVectorLayer *layer, bool selectedFeaturesOnly, const QString &baseName, const QStringList &compatibleFormats, const QString &preferredFormat, QgsProcessingContext &context, QgsProcessingFeedback *feedback, QString &layerName, long long featureLimit )
 {
   layerName.clear();
-  return convertToCompatibleFormatInternal( layer, selectedFeaturesOnly, baseName, compatibleFormats, preferredFormat, context, feedback, &layerName );
+  return convertToCompatibleFormatInternal( layer, selectedFeaturesOnly, baseName, compatibleFormats, preferredFormat, context, feedback, &layerName, featureLimit );
 }
 
 QgsFields QgsProcessingUtils::combineFields( const QgsFields &fieldsA, const QgsFields &fieldsB, const QString &fieldsBPrefix )
@@ -942,14 +1215,17 @@ QString QgsProcessingUtils::defaultRasterExtension()
 // QgsProcessingFeatureSource
 //
 
-QgsProcessingFeatureSource::QgsProcessingFeatureSource( QgsFeatureSource *originalSource, const QgsProcessingContext &context, bool ownsOriginalSource )
+QgsProcessingFeatureSource::QgsProcessingFeatureSource( QgsFeatureSource *originalSource, const QgsProcessingContext &context, bool ownsOriginalSource, long long featureLimit )
   : mSource( originalSource )
   , mOwnsSource( ownsOriginalSource )
   , mInvalidGeometryCheck( QgsWkbTypes::geometryType( mSource->wkbType() ) == QgsWkbTypes::PointGeometry
                            ? QgsFeatureRequest::GeometryNoCheck // never run geometry validity checks for point layers!
                            : context.invalidGeometryCheck() )
-  , mInvalidGeometryCallback( context.invalidGeometryCallback() )
+  , mInvalidGeometryCallback( context.invalidGeometryCallback( originalSource ) )
   , mTransformErrorCallback( context.transformErrorCallback() )
+  , mInvalidGeometryCallbackSkip( context.defaultInvalidGeometryCallbackForCheck( QgsFeatureRequest::GeometrySkipInvalid, originalSource ) )
+  , mInvalidGeometryCallbackAbort( context.defaultInvalidGeometryCallbackForCheck( QgsFeatureRequest::GeometryAbortOnInvalid, originalSource ) )
+  , mFeatureLimit( featureLimit )
 {}
 
 QgsProcessingFeatureSource::~QgsProcessingFeatureSource()
@@ -971,6 +1247,11 @@ QgsFeatureIterator QgsProcessingFeatureSource::getFeatures( const QgsFeatureRequ
     req.setInvalidGeometryCallback( mInvalidGeometryCallback );
   }
 
+  if ( mFeatureLimit != -1 && req.limit() != -1 )
+    req.setLimit( std::min( static_cast< long long >( req.limit() ), mFeatureLimit ) );
+  else if ( mFeatureLimit != -1 )
+    req.setLimit( mFeatureLimit );
+
   return mSource->getFeatures( req );
 }
 
@@ -992,6 +1273,12 @@ QgsFeatureIterator QgsProcessingFeatureSource::getFeatures( const QgsFeatureRequ
   req.setInvalidGeometryCheck( mInvalidGeometryCheck );
   req.setInvalidGeometryCallback( mInvalidGeometryCallback );
   req.setTransformErrorCallback( mTransformErrorCallback );
+
+  if ( mFeatureLimit != -1 && req.limit() != -1 )
+    req.setLimit( std::min( static_cast< long long >( req.limit() ), mFeatureLimit ) );
+  else if ( mFeatureLimit != -1 )
+    req.setLimit( mFeatureLimit );
+
   return mSource->getFeatures( req );
 }
 
@@ -1012,7 +1299,10 @@ QgsWkbTypes::Type QgsProcessingFeatureSource::wkbType() const
 
 long QgsProcessingFeatureSource::featureCount() const
 {
-  return mSource->featureCount();
+  if ( mFeatureLimit == -1 )
+    return mSource->featureCount();
+  else
+    return std::min( mFeatureLimit, static_cast< long long >( mSource->featureCount() ) );
 }
 
 QString QgsProcessingFeatureSource::sourceName() const
@@ -1062,6 +1352,26 @@ QgsExpressionContextScope *QgsProcessingFeatureSource::createExpressionContextSc
   return expressionContextScope;
 }
 
+void QgsProcessingFeatureSource::setInvalidGeometryCheck( QgsFeatureRequest::InvalidGeometryCheck method )
+{
+  mInvalidGeometryCheck = method;
+  switch ( mInvalidGeometryCheck )
+  {
+    case QgsFeatureRequest::GeometryNoCheck:
+      mInvalidGeometryCallback = nullptr;
+      break;
+
+    case QgsFeatureRequest::GeometrySkipInvalid:
+      mInvalidGeometryCallback = mInvalidGeometryCallbackSkip;
+      break;
+
+    case QgsFeatureRequest::GeometryAbortOnInvalid:
+      mInvalidGeometryCallback = mInvalidGeometryCallbackAbort;
+      break;
+
+  }
+}
+
 
 //
 // QgsProcessingFeatureSink
@@ -1082,23 +1392,41 @@ QgsProcessingFeatureSink::~QgsProcessingFeatureSink()
 bool QgsProcessingFeatureSink::addFeature( QgsFeature &feature, QgsFeatureSink::Flags flags )
 {
   bool result = QgsProxyFeatureSink::addFeature( feature, flags );
-  if ( !result )
-    mContext.feedback()->reportError( QObject::tr( "Feature could not be written to %1" ).arg( mSinkName ) );
+  if ( !result && mContext.feedback() )
+  {
+    const QString error = lastError();
+    if ( !error.isEmpty() )
+      mContext.feedback()->reportError( QObject::tr( "Feature could not be written to %1: %2" ).arg( mSinkName, error ) );
+    else
+      mContext.feedback()->reportError( QObject::tr( "Feature could not be written to %1" ).arg( mSinkName ) );
+  }
   return result;
 }
 
 bool QgsProcessingFeatureSink::addFeatures( QgsFeatureList &features, QgsFeatureSink::Flags flags )
 {
   bool result = QgsProxyFeatureSink::addFeatures( features, flags );
-  if ( !result )
-    mContext.feedback()->reportError( QObject::tr( "%1 feature(s) could not be written to %2" ).arg( features.count() ).arg( mSinkName ) );
+  if ( !result && mContext.feedback() )
+  {
+    const QString error = lastError();
+    if ( !error.isEmpty() )
+      mContext.feedback()->reportError( QObject::tr( "%1 feature(s) could not be written to %2: %3" ).arg( features.count() ).arg( mSinkName, error ) );
+    else
+      mContext.feedback()->reportError( QObject::tr( "%1 feature(s) could not be written to %2" ).arg( features.count() ).arg( mSinkName ) );
+  }
   return result;
 }
 
 bool QgsProcessingFeatureSink::addFeatures( QgsFeatureIterator &iterator, QgsFeatureSink::Flags flags )
 {
-  bool result = !QgsProxyFeatureSink::addFeatures( iterator, flags );
-  if ( !result )
-    mContext.feedback()->reportError( QObject::tr( "Features could not be written to %1" ).arg( mSinkName ) );
+  bool result = QgsProxyFeatureSink::addFeatures( iterator, flags );
+  if ( !result && mContext.feedback() )
+  {
+    const QString error = lastError();
+    if ( !error.isEmpty() )
+      mContext.feedback()->reportError( QObject::tr( "Features could not be written to %1: %2" ).arg( mSinkName, error ) );
+    else
+      mContext.feedback()->reportError( QObject::tr( "Features could not be written to %1" ).arg( mSinkName ) );
+  }
   return result;
 }
